import time
import nltk
import numpy as np
import pandas as pd
from imblearn.combine import SMOTETomek
from imblearn.over_sampling import SVMSMOTE, SMOTE, BorderlineSMOTE
from keras.callbacks import EarlyStopping
from keras.models import Sequential
from keras.layers import Dense, Embedding, SpatialDropout1D, LSTM, Dropout, Conv1D, Flatten, MaxPooling1D, ConvLSTM2D
from keras.utils import np_utils
from keras_preprocessing.sequence import pad_sequences
from keras_preprocessing.text import Tokenizer
from sklearn.feature_extraction.text import TfidfVectorizer
from sklearn.metrics import confusion_matrix, precision_score, recall_score, f1_score
from sklearn.model_selection import cross_val_score, cross_validate, cross_val_predict, RandomizedSearchCV
from sklearn.model_selection import KFold
from sklearn.preprocessing import LabelEncoder
from sklearn.pipeline import Pipeline
from nltk.corpus import stopwords
from sklearn import datasets, model_selection, preprocessing, tree
from keras.wrappers.scikit_learn import KerasClassifier
import tensorflow as tf
from keras.layers import CuDNNLSTM
from keras.callbacks import ModelCheckpoint
from imblearn.under_sampling import TomekLinks, NeighbourhoodCleaningRule, EditedNearestNeighbours
import talos as ta
import wrangle as wr
from talos.model import early_stopper
from talos.utils import live


def convertMessagesToWordEmbeddings(embeddingsPath, df):
    word_embeddings = {}
    dimensions = 25
    messages = df['message']
    # Extract pre-trained word vectors from disk
    f = open(embeddingsPath, encoding='utf-8')
    for line in f:
        values = line.split()
        word = values[0]
        coefs = np.asarray(values[1:], dtype='float32')
        word_embeddings[word] = coefs
    f.close()

    messageEmbeddings = []
    for i in messages:
        if len(i) != 0:
            v = sum([word_embeddings.get(w, np.zeros((dimensions,))) for w in i.split()]) / (len(i.split()) + 0.001)
        else:
            v = np.zeros((dimensions,))
        messageEmbeddings.append(v)

    return messageEmbeddings


# function to remove stopwords
def remove_stopwords(message, stop_words):
    new = " ".join([i for i in message if i not in stop_words])
    return new


def loadTextData(dataFile):
    df = pd.read_csv(dataFile)

    messages = []
    for s in df['message']:
        messages.append(s)

    # remove punctuations, numbers, special characters
    messages = pd.Series(messages).str.replace("[^a-zA-Z]", " ")
    messages = [str(s).lower() for s in messages]

    # remove english stopwords from the messages
    stop_words = stopwords.words('english')
    messages = [remove_stopwords(r.split(), stop_words) for r in messages]
    df['message'] = messages

    return df


# https://neurospace.io/blog/2019/04/using-talos-for-feature-hyperparameter-optimization/
def parameter_tunning(x, y, model, model_name):
    if model_name == "CNN":
        # set the parameter space for cnn
        p = {'batch_size': [5, 10, 16, 24, 32, 64, 128],
             'epochs': [100, 200, 300, 400],
             'embeddings': [10000, 20000, 30000, 40000, 50000],
             'vector_length': [50, 100, 200, 300],
             'input_lenght': [100, 150, 200, 250, 300],
             'dropout1': np.linspace(0.0, 0.6, 20, endpoint=True),
             'conv_layer1': [180, 224, 320, 352],
             'kernel_size1': [2, 3, 4],
             'activation1': ['relu', 'elu', 'tanh', 'softmax'],
             'pool_size': [2, 3, 4],
             'dropout2': np.linspace(0.0, 0.6, 20, endpoint=True),
             'conv_layer2': [32, 64, 96, 128],
             'kernel_size2': [2, 3, 4],
             'activation2': ['relu', 'elu', 'tanh', 'softmax'],
             'dropout3': np.linspace(0.0, 0.6, 20, endpoint=True),
             'activation3': ['relu', 'elu', 'tanh', 'softmax'],
             'optimizer': ['Nadam', 'adam', 'rsmprop']}
    else:
        # set the parameter space for lstm
        p = {'batch_size': [5, 10, 16, 24, 32, 64, 128],
             'epochs': [100, 200, 300, 400],
             'embeddings': [10000, 20000, 30000, 40000, 50000],
             'vector_length': [50, 100, 200, 300],
             'input_lenght': [100, 150, 200, 250, 300],
             'dropout1': np.linspace(0.0, 0.6, 20, endpoint=True),
             'lstm_layer1': [16, 32, 64, 128],
             'lstm_layer2': [50, 100, 150, 200, 250],
             'kernel_size1': [2, 3, 4],
             'activation1': ['relu', 'elu', 'tanh', 'softmax'],
             'pool_size': [2, 3, 4],
             'dropout2': np.linspace(0.0, 0.6, 20, endpoint=True),
             'conv_layer2': [32, 64, 96, 128],
             'kernel_size2': [2, 3, 4],
             'activation2': ['relu', 'elu', 'tanh', 'softmax'],
             'dropout3': np.linspace(0.0, 0.6, 20, endpoint=True),
             'activation3': ['relu', 'elu', 'tanh', 'softmax'],
             'optimizer': ['Nadam', 'adam', 'rsmprop']}

    # Find parameters
    t = ta.Scan(x=x,
                y=y,
                model=model,
                params=p,
                print_params=True,
                experiment_name='Gitter CNN',
                reduction_metric="val_loss", )

    # Cross validate evaluation of found parameters
    e = ta.Evaluate(t)
    evaluation = e.evaluate(x,
                            y,
                            model_id=None,
                            task='multiclass',
                            folds=10,
                            shuffle=True,
                            metric='val_loss',
                            asc=True)


def main():
    print("Started at: %s ", time.time())
    # tensorflow
    sess = tf.Session(config=tf.ConfigProto(log_device_placement=True))

    # loading the dataset
    datafile = '/home/PycharmProjects/MulticlassGitter/AllchannelsForClass.csv'
    dataset = loadTextData(datafile)

    print(dataset.info())

    MAX_NB_WORDS = 50000  # The maximum number of words to be used.
    MAX_SEQUENCE_LENGTH = 250  # Max number of words in each complaint.
    tokenizer = Tokenizer(num_words=MAX_NB_WORDS, filters='!"#$%&()*+,-./:;<=>?@[\]^_`{|}~', lower=True)
    tokenizer.fit_on_texts(dataset['message'].values)
    word_index = tokenizer.word_index
    print('Found %s unique words.' % len(word_index))

    print("Total messages: %s" % dataset['message'].count())
    print('category counts')
    print(dataset['category'].value_counts().sort_values(ascending=False))

    X = tokenizer.texts_to_sequences(dataset['message'].values)
    X = pad_sequences(X, maxlen=MAX_SEQUENCE_LENGTH)

    # encode class values as integers
    encoder = LabelEncoder()
    encoder.fit(dataset['category'])
    Y = encoder.transform(
        dataset['category'])  # Need to pass as array on integer to be able to compute precision, recall

    print('Shape of data tensor:', X.shape)
    print('Shape of label tensor:', Y.shape)

    print('original dataset size before resampling: ')

    # checkpoint
    filepath = "weights-improvement-{epoch:02d}-{acc:.2f}.hdf5"
    checkpoint = ModelCheckpoint(filepath, monitor='acc', verbose=1, save_best_only=True, mode='max')

    # models
    LSTM_model = KerasClassifier(build_fn=lstm_model, epochs=20, batch_size=32, verbose=1)
    CNN_model = KerasClassifier(build_fn=lstm_model, epochs=20, batch_size=32, verbose=1)

    # hyperparameter tunning
    model = CNN_model
    model_name = "CNN"
    # model = LSTM_model
    model_name = "LSTM"
    parameter_tunning(X, Y, model, model_name)

    print('Sampling experiments')
    # samplingStrategies to use
    # 'Tomek': TomekLinks(),
    # 'NCL': NeighbourhoodCleaningRule(),
    # 'SMOTE': SMOTE(),
    # 'SVMSMOTE': SVMSMOTE(),
    # 'BorderlineSMOTE': BorderlineSMOTE(),
    # 'SMOTETomek': SMOTETomek(),

    architectures = {'LSTM': LSTM_model, 'CNN': CNN_model}

    sampling_name = 'SVMSMOTE'
    algorithm = SVMSMOTE()
    sampling_experiments(architectures, sampling_name, algorithm, X, Y)
    # Need to run once per sampling technique


# Sampling experiments
def sampling_experiments(architectures, samplingName, samplingAlgorithm, X, Y):
    es = EarlyStopping(monitor='loss', mode='min', verbose=1, patience=0)
    entries = []
    results = []
    recalls = []
    precisions = []
    f1s = []
    for name, NNmodel in architectures.items():
        print(f'Model: {name} with: {samplingName}')
        cv = KFold(n_splits=10, shuffle=False)
        fold_idx = 0
        for train_index, test_index in cv.split(X):
            X_train, X_test, y_train, y_test = X[train_index], X[test_index], Y[train_index], Y[test_index]

            X_train_res, Y_train_res = samplingAlgorithm.fit_resample(X_train, y_train)
            NNmodel.fit(X_train_res, Y_train_res, epochs=4000, callbacks=[es])
            item = NNmodel.score(X_test, y_test)
            results.append(item)

            print(f'Fold {fold_idx}')  # ' - X_test = {X_test.shape}')
            y_pred = NNmodel.predict(X_test)
            p = precision_score(y_test, y_pred, average='macro')
            r = recall_score(y_test, y_pred, average='macro')
            f1 = f1_score(y_test, y_pred, average='macro')
            precisions.append(p)
            recalls.append(r)
            f1s.append(f1)

            entries.append((name, 'accuracy', fold_idx, item))
            entries.append((name, 'precision', fold_idx, p))
            entries.append((name, 'recall', fold_idx, r))
            entries.append((name, 'f1', fold_idx, f1))
            fold_idx += 1

        print('%s : %f' % ('Accuracy', np.average(results)))
        print('%s : %f' % ('Precision', np.average(precisions)))
        print('%s : %f' % ('Recall', np.average(recalls)))
        print('%s : %f' % ('F1', np.average(f1s)))

    cv_df = pd.DataFrame(entries, columns=['model_name', 'metric', 'fold_idx', 'value'])
    return cv_df


# LSTMmodel
def lstm_model():
    model = Sequential()
    model.add(
        Embedding(50000, 100, input_length=250))  # Embedding layer that uses 100 length vectors to represent each word.
    model.add(SpatialDropout1D(0.2))  # performs variational dropout in NLP models.
    model.add(CuDNNLSTM(32, return_sequences=True))  # Using keras LSTM layer is faster on GPU
    model.add(CuDNNLSTM(100))  # LSTM layer with 100 memory units. # Using keras LSTM layer is faster on GPU
    model.add(Dense(4, activation='relu'))  # The output layer must create 5 output values, one for each class.
    model.compile(loss='categorical_crossentropy', optimizer='rmsprop', metrics=[
        'accuracy'])  # Because it is a multi-class classification problem, categorical_crossentropy is used as the loss function.
    return model


def talos_lstm(x_train, y_train, x_val, y_val, params):
    model = Sequential()
    model.add(
        Embedding(params['embeddings'], params['vector_length'], input_length=params['input_length']))  # Embedding layer that uses 100 length vectors to represent each word.
    model.add(SpatialDropout1D(params['dropout1']))  # performs variational dropout in NLP models.
    model.add(CuDNNLSTM(params['lstm_layer1'], return_sequences=True))  # Using keras LSTM layer is faster on GPU
    model.add(CuDNNLSTM(params['lstm_layer2']))  # LSTM layer with 100 memory units. # Using keras LSTM layer is faster on GPU
    model.add(Dense(5, activation=params['activation1']))  # The output layer must create 5 output values, one for each class.
    model.compile(loss=params['loss'], optimizer=params['optimizer'], metrics=[
        'accuracy'])  # Because it is a multi-class classification problem, categorical_crossentropy is used as the loss function.

    history = model.fit(x_train,
                        y_train,
                        batch_size=params['batch_size'],
                        epochs=params['epochs'],
                        verbose=1,
                        validation_data=[x_val, y_val],
                        callbacks=[early_stopper(epochs=params['epochs'],
                                                 mode='moderate',
                                                 monitor='val_loss')])
    return history, model


# CNN
def cnn_model():
    model = Sequential()
    model.add(
        Embedding(50000, 100, input_length=250))  # Embedding layer that uses 100 length vectors to represent each word.
    model.add(SpatialDropout1D(0.2))  # performs variational dropout in NLP models.
    model.add(Conv1D(320, kernel_size=4, activation='relu'))
    model.add(MaxPooling1D(pool_size=2))
    model.add(Dropout(rate=1 - 0.2))
    model.add(Conv1D(128, 4, activation='tanh'))
    model.add(Dropout(rate=0.5))
    model.add(Flatten())
    model.add(Dense(4, activation='softmax'))  # The output layer must create 5 output values, one for each class.
    model.compile(loss='categorical_crossentropy', optimizer='rmsprop', metrics=[
        'accuracy'])  # Because it is a multi-class classification problem, categorical_crossentropy is used as the loss function.
    return model


def talos_cnn(x_train, y_train, x_val, y_val, params):
    model = Sequential()
    model.add(
        Embedding(params['embeddings'], params['vector_length'], input_length=params['input_length']))
    model.add(SpatialDropout1D(params['dropout1']))
    model.add(Conv1D(params['conv_layer1'], kernel_size=params['kernel_size1'], activation=params['activation1']))
    model.add(MaxPooling1D(pool_size=params['pool_size']))
    model.add(Dropout(rate=params['dropout2']))
    model.add(Conv1D(params['conv_layer2'], params['kernel_size2'], activation=params['activation2']))
    model.add(Dropout(rate=params['dropout3']))
    model.add(Flatten())
    model.add(Dense(5, activation=params['activation3']))
    model.compile(loss=params['loss'], optimizer=params['optimizer'], metrics=['accuracy'])

    history = model.fit(x_train,
                        y_train,
                        batch_size=params['batch_size'],
                        epochs=params['epochs'],
                        verbose=1,
                        validation_data=[x_val, y_val],
                        callbacks=[early_stopper(epochs=params['epochs'],
                                                 mode='moderate',
                                                 monitor='val_loss')])
    return history, model


if __name__ == "__main__":
    main()
